import asyncio
from abc import ABC, abstractmethod
from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    Generator,
    Generic,
    NamedTuple,
    Type,
    TypeVar,
    Union,
)

from langchain_core.runnables import RunnableConfig
from typing_extensions import Self, TypeGuard

if TYPE_CHECKING:
    from langgraph.pregel.types import PregelTaskDescription

V = TypeVar("V")


class ManagedValue(ABC, Generic[V]):
    def __init__(self, config: RunnableConfig) -> None:
        self.config = config

    @classmethod
    @contextmanager
    def enter(
        cls, config: RunnableConfig, **kwargs: Any
    ) -> Generator[Self, None, None]:
        try:
            value = cls(config, **kwargs)
            yield value
        finally:
            # because managed value and Pregel have reference to each other
            # let's make sure to break the reference on exit
            try:
                del value
            except UnboundLocalError:
                pass

    @classmethod
    @asynccontextmanager
    async def aenter(
        cls, config: RunnableConfig, **kwargs: Any
    ) -> AsyncGenerator[Self, None]:
        try:
            value = cls(config, **kwargs)
            yield value
        finally:
            # because managed value and Pregel have reference to each other
            # let's make sure to break the reference on exit
            try:
                del value
            except UnboundLocalError:
                pass

    @abstractmethod
    def __call__(self, step: int, task: "PregelTaskDescription") -> V:
        ...


class ConfiguredManagedValue(NamedTuple):
    cls: Type[ManagedValue]
    kwargs: dict[str, Any]


ManagedValueSpec = Union[Type[ManagedValue], ConfiguredManagedValue]

ManagedValueMapping = dict[str, ManagedValue]


def is_managed_value(value: Any) -> TypeGuard[ManagedValueSpec]:
    return (isclass(value) and issubclass(value, ManagedValue)) or isinstance(
        value, ConfiguredManagedValue
    )


@contextmanager
def ManagedValuesManager(
    values: dict[str, ManagedValueSpec],
    config: RunnableConfig,
) -> Generator[ManagedValueMapping, None, None]:
    if values:
        with ExitStack() as stack:
            yield {
                key: stack.enter_context(
                    value.cls.enter(config, **value.kwargs)
                    if isinstance(value, ConfiguredManagedValue)
                    else value.enter(config)
                )
                for key, value in values.items()
            }
    else:
        yield {}


@asynccontextmanager
async def AsyncManagedValuesManager(
    values: dict[str, ManagedValueSpec],
    config: RunnableConfig,
) -> AsyncGenerator[ManagedValueMapping, None]:
    if values:
        async with AsyncExitStack() as stack:
            # create enter tasks with reference to spec
            tasks = {
                asyncio.create_task(
                    stack.enter_async_context(
                        value.cls.aenter(config, **value.kwargs)
                        if isinstance(value, ConfiguredManagedValue)
                        else value.aenter(config)
                    )
                ): key
                for key, value in values.items()
            }
            # wait for all enter tasks
            done, _ = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
            # build mapping from spec to result
            yield {tasks[task]: task.result() for task in done}
    else:
        yield {}



from contextlib import asynccontextmanager, contextmanager
from inspect import signature
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    ContextManager,
    Generator,
    Generic,
    Optional,
    Sequence,
    Type,
    Union,
)

from langchain_core.runnables import RunnableConfig
from typing_extensions import Self

from langgraph.channels.base import BaseChannel, Value
from langgraph.errors import EmptyChannelError, InvalidUpdateError


class Context(Generic[Value], BaseChannel[Value, None, None]):
    """Exposes the value of a context manager, for the duration of an invocation.
    Context manager is entered before the first step, and exited after the last step.
    Optionally, provide an equivalent async context manager, which will be used
    instead for async invocations.

    ```python
    import httpx

    client = Channels.Context(httpx.Client, httpx.AsyncClient)
    ```
    """

    value: Value

    def __init__(
        self,
        ctx: Union[
            None, Type[ContextManager[Value]], Type[AsyncContextManager[Value]]
        ] = None,
        actx: Optional[Type[AsyncContextManager[Value]]] = None,
    ) -> None:
        if ctx is None and actx is None:
            raise ValueError("Must provide either sync or async context manager.")
        self.ctx = ctx
        self.actx = actx

    def __eq__(self, value: object) -> bool:
        return (
            isinstance(value, Context)
            and value.ctx == self.ctx
            and value.actx == self.actx
        )

    @property
    def ValueType(self) -> Any:
        """The type of the value stored in the channel."""
        return None

    @property
    def UpdateType(self) -> Type[None]:
        """The type of the update received by the channel."""
        return None

    def checkpoint(self) -> None:
        raise EmptyChannelError()

    @contextmanager
    def from_checkpoint(
        self, checkpoint: None, config: RunnableConfig
    ) -> Generator[Self, None, None]:
        if self.ctx is None:
            raise ValueError("Cannot enter sync context manager.")

        empty = self.__class__(ctx=self.ctx, actx=self.actx)
        ctx = (
            self.ctx(config)
            if signature(self.ctx).parameters.get("config")
            else self.ctx()
        )
        with ctx as value:
            empty.value = value
            yield empty

    @asynccontextmanager
    async def afrom_checkpoint(
        self, checkpoint: None, config: RunnableConfig
    ) -> AsyncGenerator[Self, None]:
        empty = self.__class__(ctx=self.ctx, actx=self.actx)
        if self.actx is not None:
            ctx = (
                self.actx(config)
                if signature(self.actx).parameters.get("config")
                else self.actx()
            )
        else:
            ctx = (
                self.ctx(config)
                if signature(self.ctx).parameters.get("config")
                else self.ctx()
            )
        if hasattr(ctx, "__aenter__"):
            async with ctx as value:
                empty.value = value
                yield empty
        else:
            with ctx as value:
                empty.value = value
                yield empty

    def update(self, values: Sequence[None]) -> bool:
        if values:
            raise InvalidUpdateError("Context channel does not accept writes.")
        return False

    def get(self) -> Value:
        try:
            return self.value
        except AttributeError:
            raise EmptyChannelError()



from contextlib import AsyncExitStack, ExitStack, asynccontextmanager, contextmanager
from typing import AsyncGenerator, Generator, Mapping

from langchain_core.runnables import RunnableConfig

from langgraph.channels.base import BaseChannel
from langgraph.checkpoint.base import Checkpoint


@contextmanager
def ChannelsManager(
    channels: Mapping[str, BaseChannel],
    checkpoint: Checkpoint,
    config: RunnableConfig,
) -> Generator[Mapping[str, BaseChannel], None, None]:
    """Manage channels for the lifetime of a Pregel invocation (multiple steps)."""
    with ExitStack() as stack:
        yield {
            k: stack.enter_context(
                v.from_checkpoint(checkpoint["channel_values"].get(k), config)
            )
            for k, v in channels.items()
        }


@asynccontextmanager
async def AsyncChannelsManager(
    channels: Mapping[str, BaseChannel],
    checkpoint: Checkpoint,
    config: RunnableConfig,
) -> AsyncGenerator[Mapping[str, BaseChannel], None]:
    """Manage channels for the lifetime of a Pregel invocation (multiple steps)."""
    async with AsyncExitStack() as stack:
        yield {
            k: await stack.enter_async_context(
                v.afrom_checkpoint(checkpoint["channel_values"].get(k), config)
            )
            for k, v in channels.items()
        }
